from django.db import models

from .lexorank import LexoRank


class RankField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("max_length", 255)
        kwargs.setdefault("insert_to_bottom", False)
        kwargs.setdefault("db_index", True)
        kwargs.setdefault("editable", False)

        self.insert_to_bottom = kwargs.pop("insert_to_bottom")
        super().__init__(*args, **kwargs)

    def pre_save(self, model_instance, add):
        current_rank = super().pre_save(model_instance, add)

        if not current_rank:
            model = model_instance._meta.model

            if model.order_with_respect_to:
                with_respect_to_kwargs = {
                    model.order_with_respect_to: getattr(
                        model_instance, model.order_with_respect_to
                    )
                }
            else:
                with_respect_to_kwargs = {}

            if self.insert_to_bottom:
                previous_rank = model.get_last_object_rank(
                    with_respect_to_kwargs=with_respect_to_kwargs
                )
                kwargs = {
                    "previous_rank": previous_rank,
                    "next_rank": None,
                }
            else:
                next_rank = model.get_first_object_rank(
                    with_respect_to_kwargs=with_respect_to_kwargs
                )
                kwargs = {
                    "previous_rank": None,
                    "next_rank": next_rank,
                }

            objects_count = model.objects.count()

            current_rank = LexoRank.get_lexorank_in_between(
                objects_count=objects_count,
                **kwargs,
            )

            model_instance.rank = current_rank

        return current_rank



from django.db import models

from .lexorank import LexoRank


class RankedModelQuerySet(models.QuerySet):
    pass


class RankedModelManager(models.Manager.from_queryset(RankedModelQuerySet)):  # type: ignore[misc] # noqa: E501
    def add_to_top(self, **kwargs):
        """Adds a new object to the top of the list."""
        if self.model.order_with_respect_to:
            with_respect_to_kwargs = {
                self.model.order_with_respect_to: kwargs[
                    self.model.order_with_respect_to
                ]
            }
        else:
            with_respect_to_kwargs = {}

        qs = self.filter(**with_respect_to_kwargs).order_by("rank")

        objects_count = qs.count()
        first_obj = qs.first()

        rank = LexoRank.get_lexorank_in_between(
            previous_rank=None,
            next_rank=first_obj.rank if first_obj else None,
            objects_count=objects_count,
        )

        return self.create(rank=rank, **kwargs)

    def add_to_bottom(self, **kwargs):
        """Adds a new object to the bottom of the list."""
        if self.model.order_with_respect_to:
            with_respect_to_kwargs = {
                self.model.order_with_respect_to: kwargs[
                    self.model.order_with_respect_to
                ]
            }
        else:
            with_respect_to_kwargs = {}

        qs = self.filter(**with_respect_to_kwargs).order_by("-rank")

        objects_count = qs.count()
        last_obj = qs.first()

        rank = LexoRank.get_lexorank_in_between(
            previous_rank=last_obj.rank if last_obj else None,
            next_rank=None,
            objects_count=objects_count,
        )

        return self.create(rank=rank, **kwargs)



from django.contrib import admin


class RankedModelAdmin(admin.ModelAdmin):
    actions = ["rebalance_ranks"]

    @admin.action(description="Rebalance ranks")
    def rebalance_ranks(modeladmin, request, queryset):
        model = queryset.first().model

        if model.order_with_respect_to:
            queryset = queryset.distinct(model.order_with_respect_to)

            for obj in queryset:
                obj.rebalance()

        else:
            queryset.first().rebalance()



from functools import cached_property
from typing import Optional

from django.contrib import admin
from django.db import models, transaction
from django.db.models import CharField
from django.db.models.functions import Length
from django.forms.models import model_to_dict

from .fields import RankField
from .lexorank import LexoRank
from .managers import RankedModelManager

CharField.register_lookup(Length, "length")


class RankedModel(models.Model):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__initial_values = model_to_dict(self)

    objects = RankedModelManager()

    rank = RankField()
    order_with_respect_to: Optional[str] = None

    class Meta:
        abstract = True
        ordering = ["rank"]

    @classmethod
    def from_db(cls, db, field_names, values):
        instance = super().from_db(db, field_names, values)
        instance._state.adding = False
        instance._state.db = db
        instance._initial_values = dict(zip(field_names, values))
        return instance

    def field_value_has_changed(self, field: str) -> bool:
        if not self.pk or not self.__initial_values:
            return False

        if getattr(self, field) != self.__initial_values[field]:
            return True

        return False

    def save(self, *args, **kwargs):
        if self.order_with_respect_to:
            if self.field_value_has_changed(self.order_with_respect_to):
                self.rank = None

        return super().save(*args, **kwargs)

    @cached_property
    def _model(self):
        return self._meta.model

    @property
    def _with_respect_to_kwargs(self):
        if not self.order_with_respect_to:
            return {}

        return {self.order_with_respect_to: getattr(self, self.order_with_respect_to)}

    @property
    def _objects_count(self):
        return self._model.objects.filter(**self._with_respect_to_kwargs).count()

    def place_on_top(self) -> "RankedModel":
        """Place object at the top of the list."""
        first_object_rank = self.get_first_object_rank(
            with_respect_to_kwargs=self._with_respect_to_kwargs
        )

        self.rank = LexoRank.get_lexorank_in_between(  # type: ignore[assignment]
            previous_rank=None,
            next_rank=first_object_rank,
            objects_count=self._objects_count,
        )

        self.save(update_fields=["rank"])

        return self

    def place_on_bottom(self) -> "RankedModel":
        """Place object at the bottom of the list."""
        last_object_rank = self.get_last_object_rank(
            with_respect_to_kwargs=self._with_respect_to_kwargs
        )

        self.rank = LexoRank.get_lexorank_in_between(  # type: ignore[assignment]
            previous_rank=last_object_rank,
            next_rank=None,
            objects_count=self._objects_count,
        )

        self.save(update_fields=["rank"])

        return self

    def place_after(self, after_obj: "RankedModel") -> "RankedModel":
        """Place object after selected one."""
        previous_rank = after_obj.rank
        next_rank = after_obj.get_next_object_rank()

        middle_rank = LexoRank.get_lexorank_in_between(
            previous_rank=previous_rank,
            next_rank=next_rank,
            objects_count=self._objects_count,
        )

        self.rank = middle_rank  # type: ignore[assignment]
        self.save(update_fields=["rank"])

        return self

    def place_before(self, before_obj: "RankedModel") -> "RankedModel":
        """Place object before selected one."""
        next_rank = before_obj.rank
        previous_rank = before_obj.get_previous_object_rank()

        middle_rank = LexoRank.get_lexorank_in_between(
            previous_rank=previous_rank,
            next_rank=next_rank,
            objects_count=self._objects_count,
        )

        self.rank = middle_rank  # type: ignore[assignment]
        self.save(update_fields=["rank"])

        return self

    def get_previous_object(self) -> Optional["RankedModel"]:
        """
        Return object that precedes provided object,
        or None if provided object is the first.
        """
        return (
            self._model.objects.filter(
                rank__lt=self.rank, **self._with_respect_to_kwargs
            )
            .order_by("-rank")
            .first()
        )

    def get_previous_object_rank(self) -> Optional[str]:
        """
        Return object rank that precedes provided object,
        or None if provided object is the first.
        """
        previous_object = self.get_previous_object()
        return previous_object.rank if previous_object else None

    def get_next_object(self) -> Optional["RankedModel"]:
        """
        Return object that follows provided object,
        or None if provided object is the last.
        """
        return (
            self._model.objects.filter(
                rank__gt=self.rank, **self._with_respect_to_kwargs
            )
            .order_by("rank")
            .first()
        )

    def get_next_object_rank(self) -> Optional[str]:
        """
        Return object rank that follows provided object,
        or None if provided object is the last.
        """
        next_object = self.get_next_object()
        return next_object.rank if next_object else None

    @transaction.atomic
    def rebalance(self) -> "RankedModel":
        """Rebalance ranks of all objects."""
        qs = (
            self._model.objects.filter(**self._with_respect_to_kwargs)
            .order_by("rank")
            .select_for_update()
        )

        objects_to_update = []

        rank = LexoRank.get_min_rank(
            objects_count=self._objects_count,
        )
        for obj in qs:
            rank = LexoRank.increment_rank(
                rank=rank,
                objects_count=self._objects_count,
            )
            obj.rank = rank
            objects_to_update.append(obj)

        self._model.objects.bulk_update(objects_to_update, ["rank"])

        self.refresh_from_db()

        return self

    @admin.display(boolean=True)
    def rebalancing_required(self) -> bool:
        """
        Return `True` if any object has rank length greater than 128, `False` otherwise.
        """
        return self._model.objects.filter(
            rank__length__gte=LexoRank.rebalancing_length,
            **self._with_respect_to_kwargs
        ).exists()

    @classmethod
    def get_first_object(cls, with_respect_to_kwargs: dict) -> Optional["RankedModel"]:
        """Return the first object if exists.."""
        if cls.order_with_respect_to and not with_respect_to_kwargs:
            raise ValueError("with_respect_to_kwargs must be provided")

        return cls.objects.filter(**with_respect_to_kwargs).order_by("rank").first()

    @classmethod
    def get_first_object_rank(cls, with_respect_to_kwargs: dict) -> Optional[str]:
        """Return the rank of the first object or None if no objects exist."""
        if cls.order_with_respect_to and not with_respect_to_kwargs:
            raise ValueError("with_respect_to_kwargs must be provided")

        first_object = cls.get_first_object(
            with_respect_to_kwargs=with_respect_to_kwargs
        )
        return first_object.rank if first_object else None

    @classmethod
    def get_last_object(cls, with_respect_to_kwargs: dict) -> Optional["RankedModel"]:
        """Return the last object if exists."""
        if cls.order_with_respect_to and not with_respect_to_kwargs:
            raise ValueError("with_respect_to_kwargs must be provided")

        return cls.objects.filter(**with_respect_to_kwargs).order_by("-rank").first()

    @classmethod
    def get_last_object_rank(cls, with_respect_to_kwargs: dict) -> Optional[str]:
        """Return the rank of the last object or None if no objects exist."""
        if cls.order_with_respect_to and not with_respect_to_kwargs:
            raise ValueError("with_respect_to_kwargs must be provided")

        last_object = cls.get_last_object(with_respect_to_kwargs=with_respect_to_kwargs)
        return last_object.rank if last_object else None



import math
import string
from typing import List, Optional, Tuple


class LexoRank:
    default_rank_length = 6
    rebalancing_length = 128
    max_rank_length = 200
    base_symbols = string.ascii_lowercase
    first_symbol = base_symbols[0]
    last_symbol = base_symbols[-1]
    base = len(base_symbols)

    @classmethod
    def char_to_int(cls, char: str) -> int:
        return ord(char) - ord(cls.first_symbol)

    @classmethod
    def int_to_char(cls, num: int) -> str:
        return chr(num + ord(cls.first_symbol))

    @classmethod
    def parse_rank(cls, rank: str) -> List[int]:
        return [cls.char_to_int(char) for char in rank]

    @classmethod
    def format_rank(cls, rank: List[int]) -> str:
        return "".join(map(cls.int_to_char, rank))

    @classmethod
    def align_ranks(cls, previous_rank: str, next_rank: str) -> Tuple[str, str]:
        max_len = max(len(previous_rank), len(next_rank))

        if max_len > cls.max_rank_length:
            raise ValueError("Rebalancing Required")

        previous_rank = previous_rank.ljust(max_len, cls.first_symbol)
        next_rank = next_rank.ljust(max_len, cls.last_symbol)

        return previous_rank, next_rank

    @classmethod
    def get_lexorank_in_between(
        cls,
        previous_rank: Optional[str],
        next_rank: Optional[str],
        objects_count: int,
        force_reorder: bool = False,
    ) -> str:
        if not previous_rank:
            previous_rank = cls.get_min_rank(objects_count=objects_count)

        if not next_rank:
            next_rank = cls.get_max_rank(objects_count=objects_count)

        previous_rank, next_rank = cls.align_ranks(previous_rank, next_rank)

        if force_reorder:
            previous_rank, next_rank = sorted([previous_rank, next_rank])
        else:
            if not previous_rank < next_rank:
                raise ValueError("Previous rank must go before than next rank.")

        previous_rank_parts = cls.parse_rank(previous_rank)
        next_rank_parts = cls.parse_rank(next_rank)

        total_diff = 0
        for i, previous_rank_part in enumerate(reversed(previous_rank_parts)):
            next_rank_part = next_rank_parts[len(next_rank_parts) - (i + 1)]
            if next_rank_part < previous_rank_part:
                next_rank_part += cls.base
                next_rank_parts[len(next_rank_parts) - (i + 1)] = next_rank_part
                next_rank_parts[len(next_rank_parts) - (i + 2)] -= 1

            diff = next_rank_part - previous_rank_part
            total_diff += diff * (cls.base**i)

        middle_rank_parts = []  # type: ignore[var-annotated]

        offset = 0
        for i, previous_rank_part in enumerate(reversed(previous_rank_parts)):
            to_add = total_diff / 2 / cls.base**i % cls.base
            middle_rank_part = previous_rank_part + to_add + offset
            offset = 0

            if middle_rank_part > cls.base:
                offset = 1
                middle_rank_part -= cls.base

            middle_rank_parts.insert(0, math.floor(middle_rank_part))

        if offset:
            middle_rank_parts.insert(0, cls.char_to_int(cls.first_symbol))

        if middle_rank_parts == previous_rank_parts:
            middle_rank_parts.append(
                (cls.char_to_int(cls.last_symbol) - cls.char_to_int(cls.first_symbol))
                // 2
            )

        return cls.format_rank(middle_rank_parts)

    @classmethod
    def get_min_rank(cls, objects_count: int) -> str:
        rank_length = cls.get_rank_length(objects_count)
        return cls.format_rank([0] * rank_length)

    @classmethod
    def get_max_rank(cls, objects_count: int) -> str:
        rank_length = cls.get_rank_length(objects_count)
        return cls.format_rank([cls.char_to_int(cls.last_symbol)] * rank_length)

    @classmethod
    def get_rank_step(cls, objects_count: int) -> int:
        rank_length = cls.get_rank_length(objects_count=objects_count)
        return int(cls.base**rank_length / objects_count - 0.5)

    @classmethod
    def get_rank_length(cls, objects_count: int) -> int:
        if objects_count == 0:
            length_required_to_place_all_objects = 1
        else:
            length_required_to_place_all_objects = math.ceil(
                math.log(objects_count, cls.base)
            )
        return min(
            cls.max_rank_length,
            max(length_required_to_place_all_objects * 2, cls.default_rank_length),
        )

    @classmethod
    def increment_rank(cls, rank: str, objects_count: int) -> str:
        step = cls.get_rank_step(objects_count=objects_count)
        rank_parts = LexoRank.parse_rank(rank)

        for i in range(len(rank_parts) - 1, -1, -1):
            if step == 0:
                break

            total = rank_parts[i] + step
            rank_parts[i] = total % cls.base
            step = total // cls.base

        if step > 0:
            rank_parts = [step] + rank_parts

        return LexoRank.format_rank(rank_parts)
